% ALIGNEIGENMAPS Align an eigenmap
%
%    [E1,E2] = ALIGNEIGENMAPS(E1,E2,np,iter,isrigid) non-rigidly aligns
%    the second eigenmap to match the first one. The number of control
%    points (default: 500) is also a downsampling parameter. The number of
%    CPD iterations (default: 50).
%
%    Example:
%       [E1,V1] = eigenmap(X1,F1);
%       [E2,V2] = eigenmap(X2,F2);
%
%       % reorder and flip signs of E2
%       [E1,E2,Q] = reordereigenmaps(X1,X2,E1,E2,V1,V2);
%
%       % aligns E2 with E1
%       [E1,E2] = aligneigenmaps(E1,E2,np);
%
%    See also eigenmap match reordereigenmaps
%
%    Reference:
%    [1] Spectral correspondence for Brain Matching,
%        Lombaert, Grady, Polimeni, Cheriet, IPMI'11

% file:    aligneigenmaps.m
% author:  Herve Lombaert
% contact: herve.lombaert@gmail.com
% date:    April 16th, 2011

function E2 = aligneigenmaps(E1,E2,np,niter,isrigid)

if nargin<3; np      = 2000;  end; % downsample using np points in embedding
if nargin<4; niter   = 50;   end; % number of CPD iterations
if nargin<5; isrigid = 0;    end; % number of CPD iterations

addpath([get_functiondir('aligneigenmaps') '/cpd']);
addpath([get_functiondir('aligneigenmaps') '/cpd/utils']);
addpath([get_functiondir('aligneigenmaps') '/cpd/mex']);
addpath([get_functiondir('aligneigenmaps') '/cpd/Nonrigid']);
addpath([get_functiondir('aligneigenmaps') '/cpd/Rigid']);
addpath([get_functiondir('aligneigenmaps') '/cpd/FGT']);

method = 'nonrigid';
if isrigid;
    method    = 'rigid';
    opt.rot   = 1;
    opt.scale = 1;
end;

%% Downsample
Y1 = E1(1:ceil(size(E1,1)/np):end,:);
Y2 = E2(1:ceil(size(E2,1)/np):end,:);

%% Register Embeddings

opt.method    = method;     % use nonrigid (default) registration
opt.beta      = 6;          % the width of Gaussian kernel (smoothness)
opt.lambda    = 50;         % regularization weight

opt.viz       = 0;          % show every iteration
opt.outliers  = 0.99;       % we downsampled, so points might not all be matching
opt.fgt       = 0;          % do not use FGT (default)
opt.normalize = 1;          % normalize to unit variance and zero mean before registering (default)
opt.corresp   = 0;          % compute correspondence vector at the end of registration (not being estimated by default)

opt.max_it    = niter;      % max number of iterations
opt.tol       = 1e-5;       % tolerance

[Transform, C] = cpd_register(Y1,Y2, opt);

%% Transform all points
if do_multithread
    if matlabpool('size')==0;  matlabpool;  end;
    
%     n = round(size(E1,1)/1000); % split the transformation into many smaller computations
%     for i=1:n
%         fprintf('%03.0f%%',100*i/n);
%         E2(i:n:end,:) = cpd_transform(E2(i:n:end,:),Transform);
%         fprintf('\b\b\b\b');
%     end
    n = matlabpool('size'); % split the transformation into many smaller computations
    parfor i=1:n
        M{i} = cpd_transform(E2(i:n:end,:),Transform);
    end
    for i=1:n
        E2(i:n:end,:) = M{i};
    end
    clear 'M';
else
    E2 = cpd_transform(E2,Transform); % no serial computation with this line
end


end
