% N = normal(X,F,D)
%   Compute normals at each nodes
%     X  - point positions
%     F  - triangles
%     N  - normal at each vertex
%     Nf - normals at each faces
%     A  - adjacency(:,2) matrix (optional as input to avoid recomputation)
%     D  - degree matrix
%
function [N,Nf,A,D] = normal(X,F,A)

    n = size(X,1);
    eps = 1e-6;
    
    % Adjacency matrix
    if nargin<3
        A = adjacency(F);
    end
    D = full(sum(A)'); % node degree
    
    % Normals on each face
    Nf = ncrossp( X(F(:,2),:)-X(F(:,1),:), X(F(:,3),:)-X(F(:,1),:) );
    
    % Average normal from triangles neighboring vertices
    N(:,1) = full(diag(sparse([F(:,1); F(:,2); F(:,3)]',[F(:,1); F(:,2); F(:,3)]',[Nf(:,1); Nf(:,1); Nf(:,1)]',n,n)));
    N(:,2) = full(diag(sparse([F(:,1); F(:,2); F(:,3)]',[F(:,1); F(:,2); F(:,3)]',[Nf(:,2); Nf(:,2); Nf(:,2)]',n,n)));
    N(:,3) = full(diag(sparse([F(:,1); F(:,2); F(:,3)]',[F(:,1); F(:,2); F(:,3)]',[Nf(:,3); Nf(:,3); Nf(:,3)]',n,n)));
    N = bsxfun(@rdivide,N,D);

    % Normalize
    d  = sqrt( sum(N.^2,2) ); d(d<eps)=1;  % normalize
    N  = bsxfun(@rdivide,N,d);
    
end

function z = ncrossp(x,y)
    % x and y are (m,3) dimensional
    eps = 1e-6;
    xnorm = sqrt(sum(x.^2,2));  xnorm(xnorm<eps)=1;  x = bsxfun(@rdivide,x,xnorm);  % normalize
    ynorm = sqrt(sum(y.^2,2));  ynorm(ynorm<eps)=1;  y = bsxfun(@rdivide,y,ynorm);  % normalize
    z = x;
    z(:,1) = x(:,2).*y(:,3) - x(:,3).*y(:,2);
    z(:,2) = x(:,3).*y(:,1) - x(:,1).*y(:,3);
    z(:,3) = x(:,1).*y(:,2) - x(:,2).*y(:,1);

    znorm = sqrt(sum(z.^2,2));  znorm(znorm<eps)=1;  z = bsxfun(@rdivide,z,znorm);  % normalize
end
