% [corr12,corr21,U1,U2,W1,W2] = spectralmatch(X1,F1,X2,F2,ki,km)
%   Spectral matching between two meshes
%     X1 - vertex positions of 1st mesh (n1 x 3)
%     F1 - triangle list containing vertex indices (ntri1 x 3)
%     X2 - vertex positions of 2nd mesh (n2 x 3)
%     F2 - triangle list of 2nd mesh (ntri2 x 3)
%     ki - number of eigenvectors to use in initial match
%     km - number of eigenvectors to use in final   match
%     corr12 - corresponding point indices from mesh 1 onto mesh 2 (n1 x 1)
%     corr21 - corresponding point indices from mesh 2 onto mesh 1 (n2 x 1)
%     U1     - weighted spectral coordinates without null vector (n1 x km-1)
%     U2     - weighted spectral coordinates without null vector (n2 x km-1)
%     W1     - adjacency matrix of mesh 1 (n1 x n1)
%     W2     - adjacency matrix of mesh 2 (n2 x n2)
%
function [corr12,corr21,U1,U2,W1,W2] = spectralmatch(X1,F1,X2,F2,ki,km)

    n1 = size(X1,1);
    n2 = size(X2,1);

    %%% Compute first sets of eigenpairs
    W1 = weightedadjacency(X1,F1);
    W2 = weightedadjacency(X2,F2);

    L1 = generallaplacian(W1);
    L2 = generallaplacian(W2);

    [E1,V1] = spectrum(L1,ki);  E1=E1(:,2:end); V1=V1(2:end);
    [E2,V2] = spectrum(L2,ki);  E2=E2(:,2:end); V2=V2(2:end);  E2 = signflip(E1,E2,X1,X2);

    E1 = weightfeatures(E1,max([V1 V2],[],2).^-.5);
    E2 = weightfeatures(E2,max([V1 V2],[],2).^-.5);

    if do_display;  displayspectrums(X1,F1,X2,F2,E1,E2);  if do_pause; pause; end;  end;

    %%% Find initial spectral matching
    [corr12,corr21] = match(E1,E2);

    if do_display;  displaycorrespondences(X1,F1,X2,F2,corr12,corr21,E1,E2);  if do_pause; pause; end;  end;

    %%% Compute connection graph eigenpairs
    W = connectivityadjacency(W1,W2,X1,X2,corr12,corr21);
    L = generallaplacian(W);

    v0 = [E1;E2];
    [E,V]   = spectrum(L,km,v0);  E=E(:,2:end); V=V(2:end);
    [U1,U2] = separatespectrum(E,n1,n2);

    if do_display;  displayspectrums(X1,F1,X2,F2,U1,U2);  if do_pause; pause; end;  end;

    %%% Find correspondence map
    U1 = weightfeatures(U1,V.^-.5);
    U2 = weightfeatures(U2,V.^-.5);
    [corr12,corr21] = match(U1,U2);

    if do_display;  displaycorrespondences(X1,F1,X2,F2,corr12,corr21,U1,U2);  if do_pause; pause; end;  end;

end