% [corr12,corr21,U1,U2,W1,W2] = spectralmatch(X1,F1,X2,F2,ki,km)
%   Spectral matching between two meshes
%     X1 - vertex positions of 1st mesh (n1 x 3)
%     F1 - triangle list containing vertex indices (ntri1 x 3)
%     X2 - vertex positions of 2nd mesh (n2 x 3)
%     F2 - triangle list of 2nd mesh (ntri2 x 3)
%     ki - number of eigenvectors to use in initial match
%     km - number of eigenvectors to use in final   match
%     corr12 - corresponding point indices from mesh 1 onto mesh 2 (n1 x 1)
%     corr21 - corresponding point indices from mesh 2 onto mesh 1 (n2 x 1)
%     U1     - weighted spectral coordinates without null vector (n1 x km-1)
%     U2     - weighted spectral coordinates without null vector (n2 x km-1)
%     W1     - adjacency matrix of mesh 1 (n1 x n1)
%     W2     - adjacency matrix of mesh 2 (n2 x n2)
%
function [corr12,corr21,U1,U2,W1,W2] = spectralmatchfeatures(X1,F1,T1,X2,F2,T2,ki,km)

    n1 = size(X1,1);
    n2 = size(X2,1);
    
    %T2 = equalize(T1,T2);

    %%% Compute first sets of eigenpairs
    %W1 = weightedadjacency(X1,F1);
    %W2 = weightedadjacency(X2,F2);
    W1 = weightedadjacencynormal(X1,F1);
    W2 = weightedadjacencynormal(X2,F2);

    L1 = generallaplacian(W1);
    L2 = generallaplacian(W2);
    
    %G1 = spdiags(exp(T1)+.001,0,speye(n1));
    %G2 = spdiags(exp(T2)+.001,0,speye(n2));
    a = 1;
    G1 = spdiags(exp(T1*a)-min(exp(T1*a))+1e-2,0,speye(n1));
    G2 = spdiags(exp(T2*a)-min(exp(T2*a))+1e-2,0,speye(n2));
    
    L1 = G1 * L1;
    L2 = G2 * L2;

    [E1,V1] = spectrum(L1,ki+1);  E1=E1(:,2:end); V1=V1(2:end);
    [E2,V2] = spectrum(L2,ki+1);  E2=E2(:,2:end); V2=V2(2:end);  E2 = signflip(E1,E2,X1,X2);

    %th = [50 20]; s = 1e-3;
    %subplot(1,2,1); trisurf(F1,X1(:,1),X1(:,2),X1(:,3),E1(:,1)); shading interp; axis equal off; view(-90,0); material dull; camlight; caxis([-1 1]*s-th(1));
    %subplot(1,2,2); trisurf(F2,X2(:,1),X2(:,2),X2(:,3),E2(:,1)); shading interp; axis equal off; view(-90,0); material dull; camlight; caxis([-1 1]*s-th(2));

    %displayspectrums(X1,F1,X2,F2,E1,E2,1);

    E1 = weightfeatures(E1,max([V1 V2],[],2).^-.5);
    E2 = weightfeatures(E2,max([V1 V2],[],2).^-.5);
    %E1 = weightfeatures(E1,V1.^-.5);
    %E2 = weightfeatures(E2,V2.^-.5);
    
    %[E1,E2] = warpdata(E1,E2,.5);
    
    %[E2,scale] = equalize(E1,E2);

    if do_display;  displayspectrums(X1,F1,X2,F2,E1,E2);  if do_pause; pause; end;  end;

    N1 = normal(X1(:,1:3),F1); % normals of mesh
    N2 = normal(X2(:,1:3),F2);
    
    Ne1 = normal(E1(:,1:3),F1); % normal of 3D spectral embeddings
    Ne2 = normal(E2(:,1:3),F2);
    
    % Prepare extra features
    Tn1 = bsxfun(@rdivide,T1,range([T1;T2])) * max(range([E1;E2]));
    Tn2 = bsxfun(@rdivide,T2,range([T1;T2])) * max(range([E1;E2]));
    %EX1 = [E1, N1, Ne1, T1];
    %EX2 = [E2, N2, Ne2, T2];
    %EX1 = [E1, N1/max(range([N1;N2]))*max(range([E1;E2])), Tn1];
    %EX2 = [E2, N2/max(range([N1;N2]))*max(range([E1;E2])), Tn2];
    EX1 = bsxfun(@times, [E1 N1 Ne1 Tn1], [ones(1,ki), [1 1 1]*.15, [1 1 1]*0.1, 2.5]);
    EX2 = bsxfun(@times, [E2 N2 Ne2 Tn2], [ones(1,ki), [1 1 1]*.15, [1 1 1]*0.1, 2.5]);
    
    [EX1,EX2] = warpdata(EX1,EX2,.5);
    
    %%% Find initial spectral matching
    [corr12,corr21] = match(EX1,EX2);

    if do_display;  displaycorr(X1,F1,X2,F2,corr12,corr21,EX1,EX2);  if do_pause; pause; end;  end;

    %U1 = E1; U2 = E2; return;
    
    %%% Compute connection graph eigenpairs
    W = connectivityadjacency(W1,W2,X1,X2,corr12,corr21);
    %W = connectivityadjacencynormal(W1,W2,X1,F1,X2,F2,corr12,corr21);
    L = generallaplacian(W);

    v0 = [E1;E2];
    [E,V]   = spectrum(L,km+1,v0);  E=E(:,2:end); V=V(2:end);
    [U1,U2] = separatespectrum(E,n1,n2);

    if do_display;  displayspectrums(X1,F1,X2,F2,U1,U2);  if do_pause; pause; end;  end;

    %%% Find correspondence map
    U1 = weightfeatures(U1,V.^-.5);
    U2 = weightfeatures(U2,V.^-.5);
    [corr12,corr21] = match(U1,U2);

    if do_display;  displaycorr(X1,F1,X2,F2,corr12,corr21,U1,U2);  if do_pause; pause; end;  end;
    
end