% [E,Lambda] = spectrum(L,ne,opts)
%    Compute the graph spectrum of L
%    L       - general graph Laplacian
%    k       - number of eigenvectors to compute, including null vector (default: 6)
%    v0      - initial eigenvector for eigs (default: [])
%    opts    - options to pass to eigs solver
%              (default: 'disp=0','isreal=1','issym=0','maxit=50')
%    E       - sorted eigenvectors in columns (n x k)
%    lambda  - sorted eigenvalues (column vector)
%
function [E,Lambda] = spectrum(L,k,v0,opts)

    %srnd(1);
    
    if nargin<4;                 opts = struct('disp',0,'isreal',1,'issym',0,'maxit',4); end;
    if nargin>2 && ~isempty(v0); v0 = sum(v0,2); opts.v0 = v0 / norm(v0);  end;

    if do_verbose;  disp(['*** Spectral decomposition of a ' num2str(size(L,1)) '^2 matrix']);  end;
    
    if isfield(opts,'srnd');  srnd(opts.srnd);  end;
    
    mytic;
    
    % Eigenvectors
    [E,Lambda,flag] = eigs(L,k,'sm',opts);
    Lambda = diag(Lambda);
    
    % Failsafe (not interested in imaginary part)
    E = real(E);
    Lambda = real(Lambda);
    
    % Sort ascending order
    [~,idx] = sort(Lambda,1,'ascend');
    Lambda = Lambda(idx);
    E = E(:,idx);
    
    if flag;  warning('Spectral decomposition did not converge');  end;
    if sum(Lambda<1e-10)>1;  warning('Multiple null vectors');  end;
    
    % Scaling
    E(:,2:end) = E(:,2:end) * diag(Lambda(2:end).^-.5);
    
    % Arbitrary signing
    signf = 1 - 2*(E(1,:)<0);  % first point in X always has positive eigenmode values
    E = bsxfun(@times,E,signf);

    mytoc;
    
end
