% Run mesh spectral matching
%
% Example of cortical brain matching using Diffeomorphic Spectral Matching
%
% Instructions:
%   Simply run it (press F5)
%   A window will invite you to find correspondences between meshes
%
% Data file format:
%   The current script runs conventional FreeSurfer mesh files
%   lh/rh.white with a list of points as (x,y,z) and triangles (f0,f1,f2)
%   lh/rh.sulc  with a list of sulcal depth (s) at each point
%   More information could be used to improve matching (see paper [1,2,3])
%   Generic meshes can be used with, or without, additional information
%
% How it works:
%   1) It first uses FOCUSR [2] to create an initial correspondence map
%   2) It then uses DiffeoSpectralMatching [1], which decomposes a
%      "correspondence graph" and produces a bijective map between cortices
%
% How to reuse code:
%   1) Point-wise correspondences (node-to-node) are in:
%      corr12 (mapped mesh 1 onto 2) and corr21 (mapped mesh 2 onto 1)
%   2) True correspondences (true location might be between nodes) are:
%      X1p = interpolate(M1,corr12,M2,W2,X2); % X1 warped onto X2
%   3) type help spectralmatchfeatures, or help spectralmatch for more help
%
% Papers to cite:
%   [1] Diffeomorphic Spectral Matching of Cortical Surfaces, IPMI 2013
%   [2] FOCUSR: A Method for Precise Surface Matching, PAMI 2012
%   [3] Fast Brain Matching with Spectral Correspondnces, IPMI 2011
%
% License:
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
%   (c) Lombaert, 2013
%
% Contact:  herve.lombaert@mail.mcgill.ca
%



%%% Clear all first
clear all;
addpath('DiffeoSpectralMatching');
addpath('freesurfer');

%%% Parameters
datapath = 'data';  % location of data (change this for your freesurfer directory)
ki       = 5;       % number of eigenvectors to find initial matching
km       = 5;       % number of eigenvectors to find connectivity graph matching
    
%%% Global variables
do_verbose      (1);
do_display      (0);
do_pause        (0);
do_multithread  (0);

%% Read FreeSurfer mesh files
id1      = 'OAS1_0001_MR1';
id2      = 'OAS1_0003_MR1';

% Mesh 1
infile = [datapath '/subjects/' id1 '/surf/lh'];
[X1,F1]   = read_surf([infile '.white']);  F1  = F1+1; % FS are 0-indexed
S1        = read_curv([infile '.sulc']);

% Mesh 2
infile = [datapath '/subjects/' id2 '/surf/lh'];
[X2,F2]   = read_surf([infile '.white']);  F2  = F2+1; % FS are 0-indexed
S2        = read_curv([infile '.sulc']);

% %% Load Meshes from Matlab files
% data = load([datapath '/surf1.mat']);  X1 = data.X;  F1 = data.F;  S1 = data.S;
% data = load([datapath '/surf2.mat']);  X2 = data.X;  F2 = data.F;  S2 = data.S;

%%%
%%% All the fun is below
%%%

%%% Spectral Correspondence
mytic;
%[corr12,corr21,U1,U2,W1,W2] = spectralmatch(X1,F1,X2,F2,ki,km);                      % Use this line for generic meshes that have no additional feature value
[corr12,corr21,U1,U2,W1,W2] = spectralmatchfeatures(X1,F1,[S1], X2,F2,[S2], ki,km);   % Use this line for brain surfaces that have additional feature values
mytoc('Total');

%%% Show Correspondences
displaycorr(X1,F1,X2,F2,corr12,corr21,U1,U2);

disp('>>> Click on the mesh to find correspondences');
subplot(121); title('[Click a point on mesh]');  subplot(122); title('[Click a point on mesh]');

return;



%%% Extra tests
displaycorr(X1,F1,X2,F2,corr12,corr21,U1(:,1),U2(:,1));
showdeformation(X1,F1,X2,F2,corr12,corr21);
showinterpolation(X1,F1,X2,F2,U1,U2,corr12,corr21,W1,W2);

